namespace GossamerVR.MorphFilterPlus
{
    using System;    
    using System.Collections;
    using System.Collections.Generic;
    using System.Text.RegularExpressions;
    using UnityEngine;
    using UnityEngine.UI;
    using UnityEngine.Events;
    using UnityEngine.EventSystems;
    
    /// <summary>
    /// Handles references and effects for a specific person and their UI elements,
    /// Persons have morph input fields, ui, and banks we track in addition to the 
    /// base save appearance button.
    /// </summary>
    public class PersonUpgrade : AtomUpgrade
    {
        public const string k_VisibleTag = "gvr_visible,";
        
        public InputField OldFieldFemale;
        public InputField OldFieldMale;
        
        public InputField NewFieldFemale;
        public InputField NewFieldMale;

        public GenerateDAZMorphsControlUI MorphUIFemale;
        public GenerateDAZMorphsControlUI MorphUIMale;

        
        // the ui panel for cloth, hair, male, female, using lookup [UIPanel, Gender]
        // except for UIPanel.Morph(4) which is handled separately.
        public GenerateDAZDynamicSelectorUI[,] UIPanels = new GenerateDAZDynamicSelectorUI[3,3];

        // user-entered tag and filter strings without our hidden tag injection, indexed by [UIPanel, Gender]
        // except for UIPanel.Morphs, which has its own thing
        //public string[,] currentBaseTags = new string[] { { "", "", "" },{ "", "", "" },{ "", "", "" } };
        public string[,] currentSearchText = new string[,] { { "", "", "" },{ "", "", "" },{ "", "", "" } };
        public Filter.RequiredSet[,] currentFilter = new Filter.RequiredSet[3,3];
        
        // we need access to the character selector to access hair and clothing.
        public DAZCharacterSelector Character;

        public DAZMorphBank[] MorphBanks;



        public PersonUpgrade(Atom owner) { this.Owner = owner; }

        // to speed up filtering and re-applying filters,
        // we cache our last search and last built filters,
        // so if the search hasn't changed we don't have to rebuild.
        private string lastAppliedStringMale;
        private Filter.RequiredSet lastAppliedFiltersMale = new Filter.RequiredSet();
        private string lastAppliedStringFemale;
        private Filter.RequiredSet lastAppliedFiltersFemale = new Filter.RequiredSet();

        protected override void OnBridgeInteract(object sender, EventArgs args)
        {
            base.OnBridgeInteract(sender, args);
            
            // when we detect input change from an input field,
            // we apply its filter.
            if (sender is InputFieldBridge)
            {
                var field = (InputFieldBridge)sender;
                if (field.NewElement.text.Length >= MorphFilterPlus.FilterThreshold || field.NewElement.text.Length == 0)
                {                    
                    this.ApplyNonMorphFilter(field);
                    var uiPanel = this.UIPanels[field.UIPanel, field.Gender];
                    if (uiPanel != null && uiPanel.tagsFilterField != null)
                    {
                        // if we haven't added our tag, do so (which auto syncs ui)
                        // otherwise just manually sync ui.
                        if (!field.IsTagField && !this.isReverting && !uiPanel.tagsFilterField.text.StartsWith(k_VisibleTag))
                        {                 
                            uiPanel.tagsFilterField.text = k_VisibleTag + uiPanel.tagsFilterField.text;
                        }  
                        else
                        {
                            
                            //Debug.Log($"calling ResyncUI(); is reverting: {this.isReverting}, current tags filter: '{uiPanel.tagsFilterField.text}', ui visible = {uiPanel.gameObject.activeInHierarchy}");
                            uiPanel.ResyncUI();
                            //if (!uiPanel.gameObject.activeInHierarchy)
                            //{
                            //    Debug.Log($"male cloth = {this.UIPanels[UIPanel.Cloth, Gender.Male].gameObject.activeInHierarchy}");
                            //    Debug.Log($"female cloth = {this.UIPanels[UIPanel.Cloth, Gender.Female].gameObject.activeInHierarchy}");
                            //    Debug.Log($"male hair = {this.UIPanels[UIPanel.Hair, Gender.Male].gameObject.activeInHierarchy}");
                            //    Debug.Log($"female hair = {this.UIPanels[UIPanel.Hair, Gender.Female].gameObject.activeInHierarchy}");
                            //}
                        }
                    }
                }
            }
        }
        

        /// <summary>
        /// An ease-of-use method allowing a fetch of the currently visible upgrade input field,
        /// primarily used for creating actions to resync or refresh the relevant UI.
        /// will hopefully be replaced soon with something more automated.
        /// </summary>        
        public InputField GetVisibleMorphField()
        {
            if (this.NewFieldMale != null && this.NewFieldMale.gameObject.activeInHierarchy)
            {
                return this.NewFieldMale;
            }
            else if (this.NewFieldFemale != null && this.NewFieldFemale.gameObject.activeInHierarchy)
            {
                return this.NewFieldFemale;
            }
            return null;
        }

        ///// <summary>
        ///// Dynamically determines which input field to apply based on 
        ///// which is visible in the scene currently, and then applies it.
        ///// </summary>
        //public void FilterMorphs(bool visibleOnly = false)
        //{
        //    if (this.NewFieldFemale != null && (!visibleOnly || this.NewFieldFemale.gameObject.activeInHierarchy))
        //    {
        //        this.FilterMorphs(this.NewFieldFemale);
        //    }
        //    else if (this.NewFieldMale != null && (!visibleOnly || this.NewFieldMale.gameObject.activeInHierarchy))
        //    {
        //        this.FilterMorphs(this.NewFieldMale);
        //    }            
        //}

        
        /// <summary>
        /// Iterates all morphs relevant to a specific field, builds filters
        /// based on the field's input, and applies visibility settings
        /// to each morph based on whether they match the filters.
        /// This is intended to be a transient effect that is restored
        /// once the UI is up-to-date so that no invisibility affects
        /// saves or other systems. This method does not perform restoration.
        /// </summary>        
        public void FilterMorphs(InputField field)
        {
            List<DAZMorphBank> banks = new List<DAZMorphBank>();

            // first we get the morph banks relevant to the field.
            string filterString;
            if (field == this.NewFieldFemale)
            {
                filterString = this.GetFilterString(this.NewFieldFemale.text);
                //if (filterString != this.lastAppliedStringFemale || forceRefresh)
                {
                    banks.Add(this.MorphBanks[MorphFilterPlus.BankIndex.FemaleChar]);
                    banks.Add(this.MorphBanks[MorphFilterPlus.BankIndex.FemaleGen]);                    
                    this.FilterMorphMethod(banks, this.NewFieldFemale, filterString, this.lastAppliedFiltersFemale);                    
                    this.lastAppliedStringFemale = filterString;
                }
            }
            else if (field == this.NewFieldMale)
            {
                filterString = this.GetFilterString(this.NewFieldMale.text);
                //if (filterString != this.lastAppliedStringMale || forceRefresh)
                {
                    banks.Add(this.MorphBanks[MorphFilterPlus.BankIndex.MaleChar]);
                    banks.Add(this.MorphBanks[MorphFilterPlus.BankIndex.MaleGen]);
                    banks.Add(this.MorphBanks[MorphFilterPlus.BankIndex.MaleGen2]);
                    this.FilterMorphMethod(banks, this.NewFieldMale, filterString, this.lastAppliedFiltersMale);
                    this.lastAppliedStringMale = filterString;
                }
            }            
        }

        /// <summary>
        /// The internal method by which morphs for a given set of banks are filtered.
        /// this method does the work of determinine if a cache should be used or building new filters,
        /// and then iterating the banks to apply the filters.
        /// </summary>        
        private void FilterMorphMethod(List<DAZMorphBank> banks, InputField field, string filterString, Filter.RequiredSet filters)
        {            
            if (string.IsNullOrEmpty(filterString.Trim()))
            {
                // if the filter is empty, we just restore all the morphs.
                this.RestoreMorphs(field, resync:false);
            }
            else if (filterString.StartsWith("@"))
            {
                // new in 3.0 starting with @ uses regular expressions.
                Regex regex = new Regex(filterString.Substring(1), RegexOptions.IgnoreCase);
                int matchCount = 0;
                foreach (DAZMorphBank bank in banks)
                {
                    foreach (DAZMorph morph in bank.morphs)
                    {

                        if (MorphFilterPlus.IgnoredMorphs.Contains(morph.uid)) continue;
                        try
                        {
                            morph.visible = regex.IsMatch(morph.lowerCaseResolvedDisplayName);
                        }
                        catch (Exception e)
                        {
                            SuperController.LogError($"MorphFilterPlus had an issue regex matching morph '{(morph == null ? "NULL" : morph.displayName)} : {e}");
                            morph.visible = true; // might as well show morphs we aren't sure about.                            
                        }
                    }
                }
            }
            else 
            {
                this.GetFilters(filterString, filters);                
                /// heads up: using continue to avoid nesting in a few places here.
                bool match = true;
                foreach (DAZMorphBank bank in banks)
                {
                    foreach (DAZMorph morph in bank.morphs)
                    {

                        if (MorphFilterPlus.IgnoredMorphs.Contains(morph.uid))
                        {
                            continue;
                        }

                        // by default everything is visible if we have no filters.
                        match = filters.Count == 0;
                        if (match)
                        {
                            morph.visible = true;
                            continue;
                        }

                        // for the required set, we presume innocence until proven guilty.
                        match = true;
                        foreach (var optionalSet in filters)
                        {
                            if (optionalSet.Count == 0) continue;

                            // for the optional set we presume guilt until proven innocent
                            match = false;
                            foreach (Filter filter in optionalSet)
                            {
                                match = filter.CompareString(morph.lowerCaseResolvedDisplayName);
                                if (match) break;
                            }

                            // once we've gone through the gamut of optional filters,
                            // if we haven't matched, the required set clearly isn't all met.
                            // so we don't need to look any more.
                            if (!match) break;
                        }

                        // whatever happened, we accept the result.
                        morph.visible = match;

                    }// end morph loop
                }// end banksy loop        
            }// end use filters condition

        }

        /// <summary>
        /// Gives us a chance to modify the raw input text before applying a filter,
        /// for example replacing spaces with include constraints.
        /// </summary>
        /// <param name="rawText"></param>
        /// <returns></returns>
        private string GetFilterString(string rawText)
        {
            return MorphFilterPlus.ConvertSpaces ? rawText.Trim().Replace(" ", "+") : rawText;
        } 

        /// <summary>
        /// This sets a reference filterlist, updating it or rebuilding it
        /// only if there is a change in the filter text. 
        /// </summary>        
        private void GetFilters(string text, Filter.RequiredSet filters)
        {
            #region the filter starts by splitting into separate requirement sets            
            // we have two tiers of "AND" (as of 3.0), the "AND" (&) and the "PLUS" (+).
            // PLUS is actually the INCLUDES/CONTAINS constraint, and is native to each filter itself.
            // AND is a top-level operation before building the filters, that separates the required filter sets.
            // AND and OR will both create logical groups of filters that will be applied.

            // First we split the input text into multiple inputs split by the AND code.
            // these will end up being filter sets with their own rules, and all of them will have
            // to give the greenlight to anything they're compared against.
            #endregion
            
            filters.Clear();
            string[] andSegments = text.Split(new[] { '&' }, StringSplitOptions.RemoveEmptyEntries);
            foreach (string segment in andSegments)
            {
                // to build a set of filters within each AND group, we first split each string into 
                // OR segments. This will give us a final set of strings that will turn into filters.
                // Since these are separated by OR, the morph only has to match one of them 
                // for us to give the greenlight.
                var optionalSet = new Filter.OptionalSet();
                string[] orSegments = segment.Split(new[] { '%', '/', '|' }, StringSplitOptions.RemoveEmptyEntries);
                foreach (var constraintString in orSegments)
                {
                    optionalSet.Add(new Filter(constraintString.ToLower()));
                }

                filters.Add(optionalSet);
            }
            
        }


        /// <summary>
        /// Unhides all morphs for this person except ignored ones.
        /// called when removing the plugin or person, as well as 
        /// before saving to ensure the morphs can be saved.
        /// </summary>
        public void RestoreMorphs(bool resync = true)
        {
            if (this.MorphBanks != null && MorphFilterPlus.IgnoredMorphs != null)
            {
                foreach (var bank in this.MorphBanks)
                {
                    this.RestoreBank(bank);
                }
            }
            if (resync) this.ResyncMorphs();
        }


        /// <summary>
        /// Restores morphs only relevant to the input field (fem or male)
        /// </summary>            
        public void RestoreMorphs(InputField field, bool resync = true)
        {
            if (this.MorphBanks != null && MorphFilterPlus.IgnoredMorphs != null)
            {
                if (field == this.NewFieldFemale)
                {
                    this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.FemaleChar]);
                    this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.FemaleGen]);
                }
                else if (field == this.NewFieldMale)
                {
                    this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.MaleChar]);
                    this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.MaleGen]);
                    this.RestoreBank(this.MorphBanks[MorphFilterPlus.BankIndex.MaleGen2]);
                }
                if (resync) this.ResyncMorphs(field);
            }
        }

        /// <summary>
        /// iterates all morphs in a bank to restore them to their pristine visibility.
        /// </summary>
        private void RestoreBank(DAZMorphBank bank)
        {
            if (bank != null)
            {
                foreach (var morph in bank.morphs)
                {
                    morph.visible = !MorphFilterPlus.IgnoredMorphs.Contains(morph.uid);
                }
            }
        }

        /// <summary>
        /// Removes upgrades to the given panels from the system.
        /// Note that at the moment Morph is not a valid panel.
        /// </summary>        
        public void RemovePanels(params int[] panels)
        {            
            // currently we only replace fields in cloth and hair.
            // this may be altered later to include things requiring unityevents
            List<IUIElementBridge> targets = new List<IUIElementBridge>();
            foreach (var bridge in this.uiBridges)
            {
                foreach (int panel in panels)
                {
                    if (bridge.UIPanel == panel)
                    {
                        targets.Add(bridge);
                        break;
                    }
                }
            }
            targets.ForEach(x => x.Remove());
        }

        /// <summary>
        /// Syncs the internal morph list of both UIs for the person.
        /// </summary>
        public void ResyncMorphs(bool visibleOnly = false)
        {
            if (this.MorphUIMale != null && (!visibleOnly||this.MorphUIMale.gameObject.activeInHierarchy)) this.MorphUIMale.ResyncMorphs();
            if (this.MorphUIFemale != null && (!visibleOnly || this.MorphUIFemale.gameObject.activeInHierarchy)) this.MorphUIFemale.ResyncMorphs();
        }

        /// <summary>
        /// Syncs the internal morph list of the ui associated with the input field.
        /// </summary>
        /// <param name="field"></param>
        public void ResyncMorphs(InputField field)
        {
            if (field == this.NewFieldFemale && this.MorphUIFemale != null) this.MorphUIFemale.ResyncMorphs();
            else if (field == this.NewFieldMale && this.MorphUIMale != null) this.MorphUIMale.ResyncMorphs();
        }

        /// <summary>
        /// Refreshes the drawn morphs on morph UI panels if they are active (or if forced)
        /// </summary>
        public void RefreshMorphUI(bool visibleOnly = false)
        {            
            if (this.MorphUIMale != null && (visibleOnly || this.MorphUIMale.gameObject.activeInHierarchy)) this.MorphUIMale.ForceCategoryRefresh();
            else if (this.MorphUIFemale != null && (visibleOnly || this.MorphUIFemale.gameObject.activeInHierarchy)) this.MorphUIFemale.ForceCategoryRefresh();
        }

        /// <summary>
        /// refreshes the drawn morphs on the page associated with a search bar,
        /// if the page is active or if forced.
        /// </summary>
        public void RefreshMorphUI(InputField field, bool forceRefresh = false)
        {
            if (field == this.NewFieldFemale && this.MorphUIFemale != null && (forceRefresh || this.MorphUIFemale.gameObject.activeInHierarchy))
            {
                this.MorphUIFemale.ForceCategoryRefresh();
            }
            else if (field == this.NewFieldMale && this.MorphUIMale != null && (forceRefresh || this.MorphUIMale.gameObject.activeInHierarchy))
            {
                this.MorphUIMale.ForceCategoryRefresh();
            }
        }
        
        /// <summary>
        /// Applies the regex or custom filter to the items associated with the field.
        /// this is very similar logically to the older apply filter method for morphs, 
        /// and there may be some shared code we can merge.
        /// </summary>        
        public void ApplyNonMorphFilter(InputFieldBridge fieldBridge)
        {
            DAZDynamicItem[] items = GetPanelItems(fieldBridge.UIPanel);
            if (fieldBridge.IsTagField)
            {
                if (fieldBridge.NewElement != null && fieldBridge.OldElement != null)
                {
                    // when editing a tag field we just prepend our visibility requirement and send it along.
                    fieldBridge.OldElement.text = k_VisibleTag + fieldBridge.NewElement.text;
                }
            }
            else if (string.IsNullOrEmpty(fieldBridge.NewElement.text))
            {
                this.RestorePanel(fieldBridge.UIPanel);
            }
            else
            {
                // if it's not a tag field, then it's a name filter field,
                // and we'll use actual filtering.                 
                if (fieldBridge.NewElement.text.StartsWith("@"))
                {
                    Regex regex = new Regex(fieldBridge.NewElement.text.Substring(1), RegexOptions.IgnoreCase);                    
                    foreach(DAZDynamicItem item in items)
                    {
                        this.SetItemVisible(item, regex.IsMatch(item.displayName));
                    }
                }
                else 
                {
                    string filterString = this.GetFilterString(fieldBridge.NewElement.text);
                    string currentString = this.currentSearchText[fieldBridge.UIPanel, fieldBridge.Gender];                    
                    if (filterString != currentString)
                    {                        
                        this.currentSearchText[fieldBridge.UIPanel, fieldBridge.Gender] = filterString;
                        Filter.RequiredSet filters = this.currentFilter[fieldBridge.UIPanel, fieldBridge.Gender];
                        if (filters == null)
                        {
                            filters = new Filter.RequiredSet();
                            this.currentFilter[fieldBridge.UIPanel, fieldBridge.Gender]= filters;
                        }

                        this.GetFilters(filterString, filters);
                        
                        bool isMatch;
                        foreach(DAZDynamicItem item in items)
                        {
                            // by default everything is visible if we have no filters.
                            isMatch = filters.Count == 0;
                            if (isMatch)
                            {
                                this.SetItemVisible(item, true);
                                continue;
                            }

                            // for the required set, we presume innocence until proven guilty.
                            isMatch = true;
                            foreach (var optionalSet in filters)
                            {
                                if (optionalSet.Count == 0) continue;

                                // for the optional set we presume guilt until proven innocent
                                isMatch = false;
                                foreach (Filter filter in optionalSet)
                                {
                                    isMatch = filter.CompareString(item.displayName.ToLower());
                                    if (isMatch) break;
                                }

                                // once we've gone through the gamut of optional filters,
                                // if we haven't matched, the required set clearly isn't all met.
                                // so we don't need to look any more.
                                if (!isMatch) break;
                            }

                            // whatever happened, we accept the result.
                            this.SetItemVisible(item, isMatch);
                        }                        
                    }
                }
            }
        }

        private void SetItemVisible(DAZDynamicItem item, bool visible)
        {
            // we apply 'visibility' here by giving or removing our special greenlight tag.
            // tags are given to the items as a comma separated string, and internally they turn into arrays.
            // so we just see if our tag is there or not.
            //Debug.Log($"setting item {item.displayName} to {visible}, current tags: {item.tags}");
            if (visible && !item.tags.StartsWith(k_VisibleTag)) item.tags = k_VisibleTag + item.tags;
            else if (!visible && item.tags.StartsWith(k_VisibleTag)) item.tags = item.tags.Substring(k_VisibleTag.Length);
            //Debug.Log($"tags after change: {item.tags}");
        }

        /// <summary>
        /// Retrieves clothing or hair based on the panel specified.
        /// note we don't need to specify a gender because there's only
        /// one 'available' ui panel and the character clothing and hair items are 
        /// automatically populated by the current gender or neutral.
        /// </summary>                
        private DAZDynamicItem[] GetPanelItems(int panel)
        {
            DAZDynamicItem[] ret_items = null;
            if (this.Character != null)
            {
                switch (panel)
                {
                    case UIPanel.Cloth:
                        ret_items = this.Character.clothingItems;
                        break;
                    case UIPanel.Hair:
                        ret_items = this.Character.hairItems;
                        break;
                }
            }
            ret_items = ret_items ?? new DAZDynamicItem[0];
            return ret_items;
        }

        public void RestorePanel(int panel)
        {
            // we hide our items by requiring a custom tag we inject during the filter.
            // when restoring, we revert to whatever the tags the user had without the injection.
            foreach (var item in this.GetPanelItems(panel))
            {
                if (item != null) 
                {
                    if (this.isReverting && item.tags.StartsWith(k_VisibleTag)) item.tags = item.tags.Substring(k_VisibleTag.Length);
                    else if (!this.isReverting && !item.tags.StartsWith(k_VisibleTag)) item.tags = k_VisibleTag + item.tags;
                }
            }
        }
        
        


        //// the following automated UI elements are added in 7.0 with the addition of cloth and hair.
        //// Morphs still use their own logic, though I will be trying to condense them in similar fashion.
        
        public void RefreshVisibleUI()
        {
            this.RefreshUIPanel(UIPanel.Morph, Gender.None, visibleOnly: true);
            this.RefreshUIPanel(UIPanel.Cloth, Gender.None, visibleOnly:true);
            this.RefreshUIPanel(UIPanel.Hair, Gender.None, visibleOnly: true);
        }


        public void RefreshUIPanel(int panel, int gender, bool visibleOnly = false)
        {
            // morphs aren't selector UI panels, for now they are using their own logic.
            if (panel == UIPanel.Morph) this.RefreshMorphUI(visibleOnly);
            else
            {                
                if (gender == Gender.None)
                {
                    this.RefreshUIPanel(this.UIPanels[panel, Gender.Male]);
                    this.RefreshUIPanel(this.UIPanels[panel, Gender.Female]);
                }
                else this.RefreshUIPanel(this.UIPanels[panel, gender]);
            }            
        }
        
        private void RefreshUIPanel(GenerateDAZDynamicSelectorUI panel, bool visibleOnly = false)
        {
            if (panel != null && (!visibleOnly || panel.gameObject.activeInHierarchy))
            {
                panel.ResyncUI();
            }
        }

        /// <summary>
        /// Destroys all generated objects, removes all modifications,
        /// unsubscribes from everything and dumps all references,
        /// in a complete reversion of the upgrade. Called when the person 
        /// or plugin is removed.
        /// </summary>
        public override void Revert()
        {
            this.isReverting = true;
            this.RestoreMorphs();
            this.RestorePanel(UIPanel.Cloth);
            this.RestorePanel(UIPanel.Hair);
            this.RefreshVisibleUI();

            base.Revert();
            
            if (this.OldFieldFemale != null) this.OldFieldFemale.gameObject.SetActive(true);
            if (this.OldFieldMale != null) this.OldFieldMale.gameObject.SetActive(true);

            if (this.NewFieldFemale != null) UnityEngine.Object.Destroy(this.NewFieldFemale.gameObject);
            if (this.NewFieldMale != null) UnityEngine.Object.Destroy(this.NewFieldMale.gameObject);

            this.UIPanels = null;
            // while not strictly necessary, we want to make sure references are released
            // to avoid memory bloat, in case VAM or Unity or a plugin maintains a reference
            // to this object somewhere, somehow.
            this.OldFieldFemale = null;
            this.OldFieldMale = null;

            this.MorphUIMale = null;
            this.MorphUIFemale = null;
            this.MorphBanks = null;
            

        }
    }

 }

/*
 *  hoping to add "NOT" functionality to tags. testing an approach 
 *  but limited control over internal tag processing, so this won't work alone, but I don't want to delete it yet
 *  because it took me awhile to write :/
 * 
 * string tagFieldText = fieldBridge.NewElement.text;
                    // first we look for any NOT filters which we are allowing for tags
                    // we do this by splitting the field by exclamations...
                    string[] notSegments = tagFieldText.Split(new[] {'!'}, StringSplitOptions.RemoveEmptyEntries);

                    // ...and then finding where the first tag in those splits end, via seeking the next comma.
                    HashSet<string> antiTags = new HashSet<string>();
                    for(int i =1; i< notSegments.Length; i++)
                    {                        
                        string endIndex = notSegments[i].IndexOf(',');
                        if (endIndex > 0)
                        {
                            Debug.Log("adding antitag: " + notSegments[i].Substring(endIndex));
                            antiTags.Add(notSegments[i].Substring(endIndex));
                        }
                        else
                        {
                            // if there's no comma, it's the last tag and we pretend there was.
                            Debug.Log("adding antitag: " + notSegments[i]);
                            antiTags.Add(notSegments[i]);
                        }
                    }

                    // now we apply a 'not' filter for anything that has the tag.
                    // however we can't just use our visibility tag, because the filter will overwrite that.
                    
                    bool match = false;
                    foreach(var item in items)
                    {
                        foreach (string antiTag in antiTags)
                        {
                            match = item.CheckMatchTag(antiTag);
                            if (match && !item.tags.EndsWith(k_AntiTag))
                            {
                                if (item.tags.StartsWith(k_VisibleTag)) item.tags = item.tags.Substring(k_VisibleTag.Length) + k_AntiTag;
                                else item.tags = item.tags + k_AntiTag;                                
                            }                    
                            else if (!match && item.tags.EndsWith(k_AntiTag))
                            {
                                item.tags = item.tags.Substring(0, k_AntiTag.Length);
                            }
                        }                        
                    }

                    // now we ensure the last item has a comma so we can reliably automate removal of the antitags                    
                    if (!tagFieldText.EndsWith(',')) tagFieldText += ',';
                    foreach (string antiTag in antiTags)
                    {
                        tagFieldText.Replace("!" + antiTag + ",", "");
                    }
*/
