namespace GossamerVR.MorphFilterPlus
{
    using System;
    using System.Collections.Generic;
    using UnityEngine;
    using UnityEngine.UI;

    /// <summary>
    /// common interface of all ui bridges,
    /// so that the main script can store them in a list.
    /// (the base class has a generic constraint)
    /// </summary>
    public interface IUIElementBridge
    {
        event EventHandler InteractEvent;
        event EventHandler RemovalEvent;
        int UIPanel { get; }
        int Gender { get; }
        void Remove();
    }
    
    /// <summary>
    /// A container representing a duplication of a button to replace
    /// an existing VAM UI button or render on top to intercept a click. 
    /// Has a public click event and can optionally forward the click on to the original button. 
    /// this is used to ensure we restore morphs before performing the normal button action.
    /// </summary>
    public abstract class UIElementBridge<T> : IUIElementBridge where T : Selectable
    {
        public event EventHandler InteractEvent;
        public event EventHandler RemovalEvent;
        public T NewElement { get; private set; }
        public T OldElement { get; private set; }
        public int UIPanel { get; private set; }
        public int Gender { get; private set; }
        protected bool canInteract => this.isOverlay || this.OldElement == null || this.OldElement.interactable;
        protected bool forwardEvents;
        
        /// <summary>
        /// if true, the new button is hidden in front of the old button.
        /// this is tracked so we don't send events if the old button is not interactable,
        /// and so we to set up the element in an invisible manner.
        /// </summary>
        protected bool isOverlay = false;
        
        protected UIElementBridge() { }

        public UIElementBridge(T oldElement, int panel, int gender, bool hideOldElement, bool forwardEvents) 
        {
            if (oldElement != null)
            {
                this.UIPanel = panel;
                this.Gender = gender;
                this.forwardEvents = forwardEvents;
                this.OldElement = oldElement;
                this.NewElement = UnityEngine.Object.Instantiate(oldElement, oldElement.transform.parent);
                this.InitNewElement();
                
                // whether or not we hide the old element let's put this one on top.
                this.NewElement.transform.SetSiblingIndex(oldElement.transform.GetSiblingIndex() + 1);

                this.isOverlay = !hideOldElement;
                if (this.isOverlay)
                {
                    this.SetupOverlay();
                }
                else
                {
                    this.OldElement.gameObject.SetActive(false);
                }
            }
        }

        /// <summary>
        /// This is where bridges should clear and add new listeners
        /// for the respective input type events.
        /// </summary>
        protected abstract void InitNewElement();

        /// <summary>
        /// For elements hidden in front of the original purely for interception,
        /// 
        /// </summary>
        protected abstract void SetupOverlay();

        protected abstract void OnRemove();

        protected void Interact()
        {
            this.InteractEvent?.Invoke(this, EventArgs.Empty);
        }
        public void Remove()
        {
            this.RemovalEvent?.Invoke(this, EventArgs.Empty);
            this.OnRemove();
            UnityEngine.Object.Destroy(this.NewElement.gameObject);
            if (this.OldElement != null)
            {
                this.OldElement.gameObject.SetActive(true);
            }
            this.OldElement = null;
            this.NewElement = null;
            this.InteractEvent = null;
            this.RemovalEvent = null;
        }
    }
}